"""
The :mod:`dariah_topics` package currently offers seven modules:

* :mod:`dariah_topics.evaluation` for evaluating semantic coherence of topics.
* :mod:`dariah_topics.postprocessing` for postprocessing text data.
* :mod:`dariah_topics.preprocessing` for preprocessing text data.
* :mod:`dariah_topics.utils` for some useful command-line utils.
* :mod:`dariah_topics.visualization` for visualizing the output of LDA models.
* :mod:`dariah_topics.modeling` is a wrapper for the LDA implementations.
"""

from dariah_topics import evaluation
from dariah_topics import postprocessing
from dariah_topics import preprocessing
from dariah_topics import utils
from dariah_topics import visualization
from dariah_topics import modeling

__author__ = "Sina Bock, Philip Duerholt, Michael Huber, Thora Hagen, Severin Simmler, Thorsten Vitt"
__version__ = "1.0"
__copyright__ = "Copyright 2017-2018, DARIAH-DE"
__credits__ = ["Sina Bock", "Philip Duerholt", "Michael Huber", "Thora Hagen", "Steffen Pielstroem", "Severin Simmler", "Thorsten Vitt"]
__maintainer__ = "Steffen Pielstroem"
__email__ = "pielstroem@biozentrum.uni-wuerzburg.de"
__status__ = "Development"



import lda
from dariah_topics import utils
import gensim

def lda(document_term_matrix, topics, iterations=1000, implementation='lda', gensim_corpus=None,
        type2id=None, path_to_mallet=None, clean_tokenized_corpus=None, document_labels=None,
        output_topic_keys=None, output_doc_topics=None, **kwargs):
    if implementation == 'lda':
        model = lda.LDA(n_topics=topics, n_iter=iterations, **kwargs)
        model.fit(document_term_matrix)
        return model
    elif implementation == 'gensim':
        model = LdaMulticore(corpus=gensim_corpus, id2word=type2id, num_topics=topics, iterations=iterations, **kwargs)
        return model
    elif implementation == 'mallet':
        Mallet = utils.Mallet(path_to_mallet)
        mallet_corpus = Mallet.import_tokenized_corpus(clean_tokenized_corpus, document_labels)
        Mallet.train_topics(mallet_corpus,
                            output_topic_keys=output_topic_keys,
                            output_doc_topics=output_topic_keys,
                            num_topics=topics,
                            num_iterations=iterations,
                            **kwargs)
    else:
        raise ValueError("{} is no supported LDA implementation".fromat(implementation))



"""
Preprocessing Text Data, Creating Matrices and Cleaning Corpora
***************************************************************

Functions of this module are for **preprocessing purpose**. You can read text \
files, `tokenize <https://en.wikipedia.org/wiki/Tokenization_(lexical_analysis)>`_ \
and segment documents (if a document is chunked into smaller segments, each segment \
counts as one document), create and read `document-term matrices <https://en.wikipedia.org/wiki/Document-term_matrix>`_, \
determine and remove features. Recurrent variable names are based on the following \
conventions:

    * ``corpus`` means an iterable containing at least one ``document``.
    * ``document`` means one single string containing all characters of a text \
    file, including whitespaces, punctuations, numbers, etc.
    * ``dkpro_document`` means a pandas DataFrame containing tokens and additional \
    information, e.g. *part-of-speech tags* or *lemmas*, produced by `DARIAH-DKPro-Wrapper <https://github.com/DARIAH-DE/DARIAH-DKPro-Wrapper>`_.
    * ``tokenized_corpus`` means an iterable containing at least one ``tokenized_document`` \
    or ``dkpro_document``.
    * ``tokenized_document`` means an iterable containing tokens of a ``document``.
    * ``document_labels`` means an iterable containing names of each ``document`` \
    and must have as much elements as ``corpus`` or ``tokenized_corpus`` does.
    * ``document_term_matrix`` means either a pandas DataFrame with rows corresponding to \
    ``document_labels`` and columns to types (distinct tokens in the corpus), whose \
    values are token frequencies, or a pandas DataFrame with a MultiIndex \
    and only one column corresponding to word frequencies. The first column of the \
    MultiIndex corresponds to a document ID (based on ``document_labels``) and the \
    second column to a type ID. The first variant is designed for small and the \
    second for large corpora.
    * ``token2id`` means a dictionary containing a token as key and an unique identifier \
    as key, e.g. ``{'first_document': 0, 'second_document': 1}``.

Contents
********
    * :func:`add_token2id` adds a token to a ``document_ids`` or ``type_ids`` dictionary \
    and assigns an unique identifier.
    * :func:`create_document_term_matrix()` creates a document-term matrix, for either \
    small or large corpora.
    * :func:`filter_pos_tags()` filters a ``dkpro_document`` by specific \
    *part-of-speech tags* and returns either tokens or, if available, lemmas.
    * :func:`find_hapax_legomena()` determines *hapax legomena* based on frequencies \
    of a ``document_term_matrix``.
    * :func:`find_stopwords()` determines *most frequent words* based on frequencies \
    of a ``document_term_matrix``.
    * :func:`read_document_term_matrix()` reads a document-term matrix from a CSV file.
    * :func:`read_from_pathlist()` reads one or multiple files based on a pathlist.
    * :func:`read_matrix_market_file()` reads a `Matrix Market <http://math.nist.gov/MatrixMarket/formats.html#MMformat>`_ \
    file for `Gensim <https://radimrehurek.com/gensim/>`_.
    * :func:`read_model()` reads a LDA model.
    * :func:`read_token2id()` reads a ``document_ids`` or ``type_ids`` dictionary \
    from a CSV file.
    * :func:`remove_features()` removes features from a ``document_term_matrix``.
    * :func:`segment()` is a wrapper for :func:`segment_fuzzy()` and segments a \
    ``tokenized_document`` into segments of a certain number of tokens, respecting existing chunks.
    * :func:`segment_fuzzy()` segments a ``tokenized_document``, tolerating existing \
    chunks (like paragraphs).
    * :func:`split_paragraphs()` splits a ``document`` or ``dkpro_document`` by paragraphs.
    * :func:`tokenize()` tokenizes a ``document`` based on a Unicode regular expression.
"""


from collections import Counter, defaultdict
import csv
from itertools import chain
from gensim.corpora import MmCorpus
import os
from lxml import etree
import numpy as np
import pandas as pd
import pickle
import regex
import logging

log = logging.getLogger('dariah_topics')


def add_token2id(token, token2id):
    """Adds token to token2id dictionary.

    With this function you can append a ``token`` to an existing ``token2id`` \
    dictionary. If ``token2id`` has *x* elements with *n* identifiers, ``token`` \
    will be element *x + 1* with identifier *n + 1*. 

    Args:
        token (str): Token.
        token2id (dict): A dictionary with tokens as keys and identifiers as values.

    Returns:
        An extended token2id dictionary.
    
    Raises:
        ValueError, if ``token`` has alread an ID in ``token2id``.
        
    Example:
        >>> token = 'example'
        >>> token2id = {'text': 0}
        >>> len(add_token2id(token, token2id)) == 2
        True
    """
    if token in token2id.keys():
        raise KeyError("{} has already an ID in token2id. Access its value with token2id[token].".format(token))
    token2id[token] = len(token2id) + 1
    return token2id


def create_document_term_matrix(tokenized_corpus, document_labels, large_corpus=False):
    """Creates a document-term matrix.

    With this function you can create a document-term-matrix where rows \
    correspond to documents in the collection and columns correspond to terms. \
    Use the function :func:`read_from_pathlist()` to read and :func:`tokenize()` \
    to tokenize your text files.

    Args:
        tokenized_corpus (list): Tokenized corpus as an iterable containing one
            or more iterables containing tokens.
        document_labels (list): Name or label of each text file.
        large_corpus (bool, optional): Set to True, if ``tokenized_corpus`` is
            very large. Defaults to False.

    Returns:
        Document-term matrix as pandas DataFrame.

    Example:
        >>> tokenized_corpus = [['this', 'is', 'document', 'one'], ['this', 'is', 'document', 'two']]
        >>> document_labels = ['document_one', 'document_two']
        >>> create_document_term_matrix(tokenized_corpus, document_labels) #doctest: +NORMALIZE_WHITESPACE
                      this   is  document  two  one
        document_one   1.0  1.0       1.0  0.0  1.0
        document_two   1.0  1.0       1.0  1.0  0.0
        >>> document_term_matrix, document_ids, type_ids = create_document_term_matrix(tokenized_corpus, document_labels, True)
        >>> isinstance(document_term_matrix, pd.DataFrame) and isinstance(document_ids, dict) and isinstance(type_ids, dict)
        True
    """
    if large_corpus:
        return _create_large_corpus_model(tokenized_corpus, document_labels)
    else:
        return _create_small_corpus_model(tokenized_corpus, document_labels)


def filter_pos_tags(dkpro_document, pos_tags=['ADJ', 'V', 'NN'], lemma=True):
    """Gets tokens or lemmas respectively of selected POS-tags from pandas DataFrame.

    With this function you can filter `DARIAH-DKPro-Wrapper <https://github.com/DARIAH-DE/DARIAH-DKPro-Wrapper>`_ \
    output. Commit a list of POS-tags to get specific tokens (if ``lemma`` False) \
    or lemmas (if ``lemma`` True). 
    Use the function :func:`read_from_pathlist()` to read CSV files.

    Args:
        dkpro_document (pandas.DataFrame): DARIAH-DKPro-Wrapper output.
        pos_tags (list, optional): List of desired POS-tags. Defaults
            to ``['ADJ', 'V', 'NN']``.
        lemma (bool, optional): If True, lemmas will be selected, otherwise tokens.
            Defaults to True.

    Yields:
        A pandas DataFrame containing tokens or lemmas.

    Example:
        >>> dkpro_document = pd.DataFrame({'CPOS': ['ART', 'V', 'ART', 'NN'],
        ...                                'Token': ['this', 'was', 'a', 'document'],
        ...                                'Lemma': ['this', 'is', 'a', 'document']})
        >>> list(filter_pos_tags(dkpro_document)) #doctest: +NORMALIZE_WHITESPACE
        [1          is
        3    document
        Name: Lemma, dtype: object]
    """
    tokenized_document = dkpro_document[dkpro_document['CPOS'].isin(pos_tags)]
    if lemma:
        log.info("Selecting {} lemmas ...".format(pos_tags))
        yield tokenized_document['Lemma']
    else:
        log.info("Selecting {} tokens ...".format(pos_tags))
        yield tokenized_document['Token']


def find_hapax_legomena(document_term_matrix, type_ids=None):
    """Creates a list with hapax legommena.

    With this function you can determine *hapax legomena* for each document. \
    Use the function :func:`create_document_term_matrix()` to create a \
    document-term matrix.

    Args:
        document_term_matrix (pandas.DataFrame): A document-term matrix.
        type_ids (dict): A dictionary with types as key and identifiers as values.
            If ``document_term_matrix`` is designed for large corpora, you have
            to commit ``type_ids``, too.

    Returns:
        Hapax legomena in a list.

    Example:
        >>> document_labels = ['document']
        >>> tokenized_corpus = [['hapax', 'stopword', 'stopword']]
        >>> document_term_matrix = create_document_term_matrix(tokenized_corpus, document_labels)
        >>> find_hapax_legomena(document_term_matrix)
        ['hapax']
        >>> document_term_matrix, _, type_ids = create_document_term_matrix(tokenized_corpus, document_labels, large_corpus=True)
        >>> find_hapax_legomena(document_term_matrix, type_ids)
        ['hapax']
    """
    log.info("Determining hapax legomena ...")
    if isinstance(document_term_matrix.index, pd.MultiIndex):
        log.debug("Large corpus model ...")
        return _hapax_legomena_large_corpus_model(document_term_matrix, type_ids)
    else:
        log.debug("Small corpus model ...")
        return document_term_matrix.loc[:, document_term_matrix.max() == 1].columns.tolist()


def find_stopwords(document_term_matrix, most_frequent_tokens=100, type_ids=None):
    """Creates a list with stopword based on most frequent tokens.

    With this function you can determine *most frequent tokens*, also known as \
    *stopwords*. First, you have to translate your corpus into a document-term \
    matrix.
    Use the function :func:`create_document_term_matrix()` to create a \
    document-term matrix.

    Args:
        document_term_matrix (pandas.DataFrame): A document-term matrix.
        most_frequent_tokens (int, optional): Treshold for most frequent tokens.
        type_ids (dict): If ``document_term_matrix`` is designed for large corpora,
            you have to commit ``type_ids``, too.

    Returns:
        Most frequent tokens in a list.

    Example:
        >>> document_labels = ['document']
        >>> tokenized_corpus = [['hapax', 'stopword', 'stopword']]
        >>> document_term_matrix = create_document_term_matrix(tokenized_corpus, document_labels)
        >>> find_stopwords(document_term_matrix, 1)
        ['stopword']
        >>> document_term_matrix, _, type_ids = create_document_term_matrix(tokenized_corpus, document_labels, large_corpus=True)
        >>> find_stopwords(document_term_matrix, 1, type_ids)
        ['stopword']
    """
    log.info("Determining stopwords ...")
    if isinstance(document_term_matrix.index, pd.MultiIndex):
        log.debug("Large corpus model ...")
        return _stopwords_large_corpus_model(document_term_matrix, type_ids, most_frequent_tokens)
    else:
        log.debug("Small corpus model ...")
        return document_term_matrix.iloc[:, :most_frequent_tokens].columns.tolist()


def read_document_term_matrix(filepath):
    """Reads a document-term matrix from CSV file.

    With this function you can read a CSV file containing a document-term \
    matrix.
    Use the function :func:`create_document_term_matrix()` to create a document-term \
    matrix.

    Args:
        filepath (str): Path to CSV file.

    Returns:
        A document-term matrix as pandas DataFrame.
    
    Example:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(suffix='.csv') as tmpfile:
        ...     tmpfile.write(b',this,is,an,example,text\\ndocument,1,0,1,0,1') and True
        ...     tmpfile.flush()
        ...     read_document_term_matrix(tmpfile.name) #doctest: +NORMALIZE_WHITESPACE
        True
                  this  is  an  example  text
        document     1   0   1        0     1
        >>> with tempfile.NamedTemporaryFile(suffix='.csv') as tmpfile:
        ...     tmpfile.write(b'document_id,type_id,0\\n1,1,1') and True
        ...     tmpfile.flush()
        ...     read_document_term_matrix(tmpfile.name) #doctest: +NORMALIZE_WHITESPACE
        True
                             0
        document_id type_id   
        1           1        1
    """
    document_term_matrix = pd.read_csv(filepath)
    if 'document_id' and 'type_id' in document_term_matrix:
        return document_term_matrix.set_index(['document_id', 'type_id'])
    else:
        document_term_matrix = document_term_matrix.set_index('Unnamed: 0')
        document_term_matrix.index.name = None
        return document_term_matrix


def read_from_pathlist(pathlist, file_format=None, xpath_expression='//tei:text', sep='\t', csv_columns=None):
    """Reads text files based on a pathlist.

    With this function you can read multiple file formats:
        * Plain text files (``.txt``).
        * TEI XML files (``.xml``).
        * CSV files (``.csv``), e.g. produced by `DARIAH-DKPro-Wrapper <https://github.com/DARIAH-DE/DARIAH-DKPro-Wrapper>`_. 

    The argument ``pathlist`` is an iterable of full or relative paths. In case of \
    CSV files, you have the ability to select specific columns via ``columns``. \
    If there are multiple file formats in ``pathlist``, do not specify ``file_format`` \
    and file extensions will be considered.

    Args:
        pathlist (list): One or more paths to text files.
        file_format (str, optional): Format of the files. Possible values are
            ``text``, ``xml`` and ``csv`. If None, file extensions will be considered.
            Defaults to None.
        xpath_expression (str, optional): XPath expressions to match part of the
            XML file. Defaults to ``//tei:text``.
        sep (str, optional): Separator of CSV file. Defaults to ``'\\t'``
        columns (list, optional): Column name or names for CSV files. If None, the
            whole file will be processed. Defaults to None.

    Yields:
        A ``document`` as str or, in case of a CSV file, a ``dkpro_document`` as a pandas DataFrame.

    Raises:
        ValueError, if ``file_format`` is not supported.

    Example:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(suffix='.txt') as first:
        ...     pathlist = []
        ...     first.write(b"This is the first example.") and True
        ...     first.flush()
        ...     pathlist.append(first.name)
        ...     with tempfile.NamedTemporaryFile(suffix='.txt') as second:
        ...         second.write(b"This is the second example.") and True
        ...         second.flush()
        ...         pathlist.append(second.name)
        ...         list(read_from_pathlist(pathlist, 'text'))
        True
        True
        ['This is the first example.', 'This is the second example.']
    """
    log.info("Reading {} files ...".format(len(pathlist)))
    for n, file in enumerate(pathlist):
        log.debug("File #{}".format(n))
        _, extension = os.path.splitext(file)
        if file_format == 'text' or extension == '.txt':
            yield _read_txt(file)
        elif file_format == 'xml' or extension == '.xml':
            yield _read_xml(file, xpath_expression)
        elif file_format == 'csv' or extension == '.csv':
            yield _read_csv(file, sep, csv_columns)
        else:
            if file_format is None:
                log.error("Skipping {}, because the file format {} is not supported.".format(file, extension))
                pass
            else:
                raise ValueError("Unable to read {}, because the file format {} is not supported.".format(file, file_format))


def read_matrix_market_file(filepath):
    """Reads a Matrix Market file for Gensim.

    With this function you can read a Matrix Market file to process it with \
    `Gensim <https://radimrehurek.com/gensim/>`_.

    Args:
        filepath (str): Path to Matrix Market file.

    Returns:
        Matrix Market model for Gensim.
    """
    if os.path.splitext(filepath)[1] != '.mm':
        raise ValueError("The file {} is not a Matrix Market file.".format(filepath))
    return MmCorpus(filepath)


def read_model(filepath):
    """Reads a LDA model.

    With this function you can read a LDA model, if it was saved using :module:`pickle`.
    If you want to read MALLET models, you have to specify a parameter of the
    function :func:`create_mallet_model()`.

    Args:
        filepath (str): Path to LDA model, e.g. ``/home/models/model.pickle``.

    Returns:
        A LDA model.

    Example:
        >>> import lda
        >>> import gensim
        >>> import tempfile
        >>> a = lda.LDA
        >>> with tempfile.NamedTemporaryFile(suffix='.pickle') as tmpfile:
        ...     pickle.dump(a, tmpfile, protocol=pickle.HIGHEST_PROTOCOL)
        ...     tmpfile.flush()
        ...     read_model(tmpfile.name) == a
        True
        >>> a = gensim.models.LdaModel
        >>> with tempfile.NamedTemporaryFile(suffix='.pickle') as tmpfile:
        ...     pickle.dump(a, tmpfile, protocol=pickle.HIGHEST_PROTOCOL)
        ...     tmpfile.flush()
        ...     read_model(tmpfile.name) == a
        True
    """
    with open(filepath, 'rb') as model:
        return pickle.load(model)


def read_token2id(filepath):
    """Reads a token2id dictionary from CSV file.

    With this function you can read a CSV-file containing a document or type dictionary.

    Args:
        filepath (str): Path to CSV file.

    Returns:
        A dictionary.
    
    Example:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(suffix='.csv') as tmpfile:
        ...     tmpfile.write(b"0,this\\n1,is\\n2,an\\n3,example") and True
        ...     tmpfile.flush()
        ...     read_token2id(tmpfile.name)
        True
        {0: 'this', 1: 'is', 2: 'an', 3: 'example'}
    """
    dictionary = pd.read_csv(filepath, header=None)
    dictionary.index = dictionary[0]
    dictionary = dictionary[1]
    return dictionary.to_dict()
    
    
def remove_features(features, document_term_matrix=None, tokenized_corpus=None, type_ids=None):
    """Removes features based on a list of tokens.

    With this function you can clean your corpus (either a document-term matrix \
    or a ``tokenized_corpus``) from *stopwords* and *hapax legomena*.
    Use the function :func:`create_document_term_matrix()` or :func:`tokenize` to \
    create a document-term matrix or to tokenize your corpus, respectively.

    Args:
        features (list): A list of tokens.
        document_term_matrix (pandas.DataFrame, optional): A document-term matrix.
        tokenized_corpus (list, optional): An iterable of one or more ``tokenized_document``.
        type_ids (dict, optional): A dictionary with types as key and identifiers as values.

    Returns:
        A clean document-term matrix as pandas DataFrame or ``tokenized_corpus`` as list.

    Example:
        >>> document_labels = ['document']
        >>> tokenized_corpus = [['this', 'is', 'a', 'document']]
        >>> document_term_matrix = create_document_term_matrix(tokenized_corpus, document_labels)
        >>> features = ['this']
        >>> remove_features(features, document_term_matrix) #doctest: +NORMALIZE_WHITESPACE
                   is  document    a
        document  1.0       1.0  1.0
        >>> document_term_matrix, _, type_ids = create_document_term_matrix(tokenized_corpus, document_labels, large_corpus=True)
        >>> len(remove_features(features, document_term_matrix, type_ids=type_ids))
        3
        >>> list(remove_features(features, tokenized_corpus=tokenized_corpus))
        [['is', 'a', 'document']]
    """
    log.info("Removing features ...")
    if document_term_matrix is not None and tokenized_corpus is None:
        if isinstance(document_term_matrix.index, pd.MultiIndex):
            return _remove_features_from_large_corpus_model(document_term_matrix, type_ids, features)
        else:
            return _remove_features_from_small_corpus_model(document_term_matrix, features)
    elif document_term_matrix is None and tokenized_corpus is not None:
        clean_tokenized_corpus = pd.Series() # schöner machen
        for n, tokenized_document in enumerate(tokenized_corpus):
            clean_tokenized_corpus[str(n)] = _remove_features_from_tokenized_document(tokenized_document, features)
        return clean_tokenized_corpus
    else:
        raise ValueError("Commit either document-term matrix or tokenized_corpus.")


def segment(document, segment_size=1000, tolerance=0, chunker=None,
            tokenizer=None, flatten_chunks=True, materialize=True):
    """Segments a document into segments of about ``segment_size`` tokens, respecting existing chunks.

    Consider you have a document. You wish to split the document into \
    segments of about 1000 tokens, but you prefer to keep paragraphs together \
    if this does not increase or decrease the token size by more than 5%.
    This is a convenience wrapper around :func:`segment_fuzzy()`.

    Args:
        document (list): The document to process. This is an iterable of
            chunks, each of which is an iterable of tokens.
        segment_size (int): The target size of each segment, in tokens. Defaults
            to 1000.
        tolerance (float, optional): How much may the actual segment size differ from
            the segment_size? If ``0 < tolerance < 1``, this is interpreted as a
            fraction of the segment_size, otherwise it is interpreted as an
            absolute number. If ``tolerance < 0``, chunks are never split apart.
            Defaults to None.
        chunker (callable, optional): A one-argument function that cuts the document into
            chunks. If this is present, it is called on the given document.
            Defaults to None.
        tokenizer (callable, optional): A one-argument function that tokenizes each chunk.
            Defaults to None.
        flatten_chunks (bool, optional): If True, undo the effect of the chunker by
            chaining the chunks in each segment, thus each segment consists of
            tokens. This can also be a one-argument function in order to
            customize the un-chunking. Defaults to True.
        materialize (bool, optional): If True, materializes the segments. Defaults to True.

    Example:
        >>> segment([['This', 'is', 'the', 'first', 'chunk'],
        ...          ['this', 'is', 'the', 'second', 'chunk']], 2) #doctest: +NORMALIZE_WHITESPACE
        [['This', 'is'],
        ['the', 'first'],
        ['chunk', 'this'],
        ['is', 'the'],
        ['second', 'chunk']]
    """
    if chunker is not None:
        document = chunker(document)
    if tokenizer is not None:
        document = map(tokenizer, document)

    segments = segment_fuzzy(document, segment_size, tolerance)

    if flatten_chunks:
        if not callable(flatten_chunks):
            def flatten_chunks(segment):
                return list(chain.from_iterable(segment))
        segments = map(flatten_chunks, segments)
    if materialize:
        segments = list(segments)
    return segments


def segment_fuzzy(document, segment_size=5000, tolerance=0.05):
    """Segments a document, tolerating existing chunks (like paragraphs).

    Consider you have a ``document``. You wish to split the ``document`` into \
    segments of about 1000 tokens, but you prefer to keep paragraphs together \
    if this does not increase or decrease the token size by more than 5%.

    Args:
        document (list): The document to process. This is an iterable of
            chunks, each of which is an iterable of tokens.
        segment_size (int, optional): The target length of each segment in tokens.
            Defaults to 5000.
        tolerance (float, optional): How much may the actual segment size differ from
            the ``segment_size``? If ``0 < tolerance < 1``, this is interpreted as a
            fraction of the ``segment_size``, otherwise it is interpreted as an
            absolute number. If ``tolerance < 0``, chunks are never split apart.
            Defaults to 0.05.

    Yields:
        Segments. Each segment is a list of chunks, each chunk is a list of
        tokens.

    Example:
        >>> list(segment_fuzzy([['This', 'is', 'the', 'first', 'chunk'],
        ...                     ['this', 'is', 'the', 'second', 'chunk']], 2)) #doctest: +NORMALIZE_WHITESPACE
        [[['This', 'is']],
        [['the', 'first']],
        [['chunk'], ['this']],
        [['is', 'the']],
        [['second', 'chunk']]]
    """
    if tolerance > 0 and tolerance < 1:
        tolerance = round(segment_size * tolerance)

    current_segment = []
    current_size = 0
    carry = None
    doc_iter = iter(document)

    try:
        while True:
            chunk = list(carry if carry else next(doc_iter))
            carry = None
            current_segment.append(chunk)
            current_size += len(chunk)

            if current_size >= segment_size:
                too_long = current_size - segment_size
                too_short = segment_size - (current_size - len(chunk))

                if tolerance >= 0 and min(too_long, too_short) > tolerance:
                    chunk_part0 = chunk[:-too_long]
                    carry = chunk[-too_long:]
                    current_segment[-1] = chunk_part0
                elif too_long >= too_short:
                    carry = current_segment.pop()
                yield current_segment
                current_segment = []
                current_size = 0
    except StopIteration:
        pass

    if current_segment:
        yield current_segment


def split_paragraphs(document, sep=regex.compile(r'\n')):
    """Splits the given document by paragraphs.

    With this function you can split a document by paragraphs. In case of a \
    document as str, you also have the ability to select a certain regular \
    expression to split the document.
    Use the function :func:`read_from_pathlist()` to read files.

    Args:
        document Union(str, pandas.DataFrame): Document text or DARIAH-DKPro-Wrapper output.
        sep (regex.Regex, optional): Separator indicating a paragraph.

    Returns:
        A list of paragraphs.

    Example:
        >>> document = "First paragraph\\nsecond paragraph."
        >>> split_paragraphs(document)
        ['First paragraph', 'second paragraph.']
        >>> dkpro_document = pd.DataFrame({'Token': ['first', 'paragraph', 'second', 'paragraph', '.'],
        ...                                'ParagraphId': [1, 1, 2, 2, 2]})
        >>> split_paragraphs(dkpro_document)[0] #doctest: +NORMALIZE_WHITESPACE
                         Token
        ParagraphId           
        1                first
        1            paragraph
    """
    if isinstance(document, str):
        if not hasattr(sep, 'match'):
            sep = regex.compile(sep)
        splitted_document = sep.split(document)
        return list(filter(str.strip, splitted_document)) #remove elements containing only whitespaces
    elif isinstance(document, pd.DataFrame):
        grouped_document = document.set_index('ParagraphId').groupby(level=0)
        return [paragraphs for _, paragraphs in grouped_document]


def tokenize(document, pattern=r'\p{L}+\p{P}?\p{L}+', lower=True):
    """Tokenizes with Unicode regular expressions.

    With this function you can tokenize a ``document`` with a regular expression. \
    You also have the ability to commit your own regular expression. The default \
    expression is ``\p{Letter}+\p{Punctuation}?\p{Letter}+``, which means one or \
    more letters, followed by one or no punctuation, followed by one or more \
    letters. So, one letter words will not match. In case you want to lower \
    all tokens, set the argument ``lower`` to True (it is by default).    
    Use the functions :func:`read_from_pathlist()` to read your text files.

    Args:
        document (str): Document text.
        pattern (str, optional): Regular expression to match tokens.
        lower (boolean, optional): If True, lowers all characters. Defaults to True.

    Yields:
        All matching tokens in the ``document``.

    Example:
        >>> list(tokenize("This is 1 example text."))
        ['this', 'is', 'example', 'text']
    """
    log.debug("Tokenizing document ...")
    if lower:
        log.debug("Lowering all characters ...")
        document = document.lower()
    compiled_pattern = regex.compile(pattern)
    tokenized_document = compiled_pattern.finditer(document)
    for match in tokenized_document:
        yield match.group()


def _create_bag_of_words(document_labels, tokenized_corpus):
    """Creates a bag-of-words model.

    This private function is wrapped in :func:`_create_large_corpus_model()`. The \
    first level consists of the document label as key, and the dictionary \
    of counts as value. The second level consists of type ID as key, and the \
    count of types in document pairs as value.

    Args:
        document_labels (list): Iterable of document labels.
        tokenized_corpus (list): Tokenized corpus as an iterable
            containing one or more iterables containing tokens.

    Returns:
        A bag-of-words model as dictionary of dictionaries, document IDs and type IDs.

    Example:
        >>> document_labels = ['exampletext']
        >>> tokenized_corpus = [['this', 'is', 'an', 'example', 'text']]
        >>> bag_of_words, document_ids, type_ids = _create_bag_of_words(document_labels, tokenized_corpus)
        >>> isinstance(bag_of_words, dict) and isinstance(document_ids, dict) and isinstance(type_ids, dict)
        True
    """
    document_ids = _token2id(document_labels)
    type_ids = _token2id(tokenized_corpus)
    bag_of_words = defaultdict(dict)
    for document_label, tokenized_document in zip(document_labels, tokenized_corpus):
        bag_of_words[document_label] = Counter([type_ids[token] for token in tokenized_document])
    return {document_ids[id_]: doc for id_, doc in bag_of_words.items()}, document_ids, type_ids


def _create_large_corpus_model(tokenized_corpus, document_labels):
    """Creates a document-term matrix for large corpora.

    This private function is wrapped in :func:`create_document_term_matrix()` and \
    creates a pandas DataFrame containing document and type IDs as MultiIndex \
    and type frequencies as values representing the counts of tokens for each \
    token in each document.

    Args:
        tokenized_corpus (list): Tokenized corpus as an iterable
            containing one or more iterables containing tokens.
        document_labels (list): Iterable of document labels.

    Returns:
        A document-term matrix as pandas DataFrame, ``document_ids`` and ``type_ids``.

    Todo:
        * Make the whole function faster.

    Example:
        >>> tokenized_corpus = [['this', 'is', 'document', 'one'], ['this', 'is', 'document', 'two']]
        >>> document_labels = ['document_one', 'document_two']
        >>> document_term_matrix, document_ids, type_ids = _create_large_corpus_model(tokenized_corpus, document_labels)
        >>> isinstance(document_term_matrix, pd.DataFrame) and isinstance(document_ids, dict) and isinstance(type_ids, dict)
        True
    """
    bag_of_words, document_ids, type_ids = _create_bag_of_words(document_labels, tokenized_corpus)
    multi_index = _create_multi_index(bag_of_words)
    document_term_matrix = pd.DataFrame(np.zeros((len(multi_index), 1), dtype=int), index=multi_index)
    index_iterator = multi_index.groupby(multi_index.get_level_values('document_id'))

    for document_id in range(1, len(multi_index.levels[0]) + 1):
        for type_id in [val[1] for val in index_iterator[document_id]]:
            document_term_matrix.set_value((document_id, type_id), 0, int(bag_of_words[document_id][type_id]))
    return document_term_matrix, document_ids, type_ids


def _create_multi_index(bag_of_words):
    """Creates a MultiIndex for a pandas DataFrame.

    This private function is wrapped in :func:`_create_large_corpus_model()`.

    Args:
        bag_of_words (dict): A bag-of-words model of ``{document_id: {type_id: frequency}}``.

    Returns:
        Pandas MultiIndex.

    Example:
        >>> bag_of_words = {1: {1: 2, 2: 3, 3: 4}}
        >>> _create_multi_index(bag_of_words)
        MultiIndex(levels=[[1], [1, 2, 3]],
                   labels=[[0, 0, 0], [0, 1, 2]],
                   names=['document_id', 'type_id'])
    """
    tuples = []
    for document_id in range(1, len(bag_of_words) + 1):
        if len(bag_of_words[document_id]) == 0:
            tuples.append((document_id, 0))
        for type_id in bag_of_words[document_id]:
            tuples.append((document_id, type_id))
    return pd.MultiIndex.from_tuples(tuples, names=['document_id', 'type_id'])


def _create_small_corpus_model(tokenized_corpus, document_labels):
    """Creates a document-term matrix for small corpora.

    This private function is wrapped in :func:`create_document_term_matrix()`.

    Args:
        tokenized_corpus (list): Tokenized corpus as an iterable
            containing one or more iterables containing tokens.
        document_labels (list): Name or label of each text file.


    Returns:
        Document-term matrix as pandas DataFrame.

    Example:
        >>> tokenized_corpus = [['this', 'is', 'document', 'one'], ['this', 'is', 'document', 'two']]
        >>> document_labels = ['document_one', 'document_two']
        >>> _create_small_corpus_model(tokenized_corpus, document_labels) #doctest: +NORMALIZE_WHITESPACE
                      this   is  document  two  one
        document_one   1.0  1.0       1.0  0.0  1.0
        document_two   1.0  1.0       1.0  1.0  0.0
    """
    log.info("Creating document-term matrix for small corpus ...")
    document_term_matrix = pd.DataFrame()
    for tokenized_document, document_label in zip(tokenized_corpus, document_labels):
        log.debug("Updating {} in document-term matrix ...".format(document_label))
        current_document = pd.Series(Counter(tokenized_document))
        current_document.name = document_label
        document_term_matrix = document_term_matrix.append(current_document)
    document_term_matrix = document_term_matrix.loc[:, document_term_matrix.sum().sort_values(ascending=False).index]
    return document_term_matrix.fillna(0)


def _hapax_legomena_large_corpus_model(document_term_matrix, type_ids):
    """Determines hapax legomena in large corpus model.

    This private function is wrapped in :func:`find_hapax_legomena()`.

    Args:
        document_term_matrix (pandas.DataFrame): A document-term matrix.
        type_ids (dict): A dictionary with types as key and identifiers as values.

    Returns:
        Hapax legomena in a list.

    Example:
        >>> document_labels = ['document']
        >>> tokenized_corpus = [['hapax', 'stopword', 'stopword']]
        >>> document_term_matrix, _, type_ids = create_document_term_matrix(tokenized_corpus, document_labels, large_corpus=True)
        >>> find_hapax_legomena(document_term_matrix, type_ids)
        ['hapax']
    """
    id2type = {id_: type_ for type_, id_ in type_ids.items()}
    document_term_matrix_collapsed = document_term_matrix.groupby(document_term_matrix.index.get_level_values('type_id')).sum()
    hapax_legomena = document_term_matrix_collapsed.loc[document_term_matrix_collapsed[0] == 1]
    return [id2type[token] for token in hapax_legomena.index.get_level_values('type_id')]


def _read_csv(filepath, sep, columns):
    """Reads a CSV file based on its path.
    
    This private function is wrapped in `read_from_pathlist()`.
    
    Args:
        filepath (str): Path to CSV file.
        sep (str): Separator of CSV file.
        columns (list): Column names for the CSV file. If None, the whole file will be processed.

    Returns:
        A ``dkpro_document`` as pandas DataFrame with additional information,
            e.g. lemmas or POS-tags.
    
    Example:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(suffix='.csv') as tmpfile:
        ...     tmpfile.write(b"Token,POS\\nThis,ART\\nis,V\\na,ART\\nCSV,NN\\nexample,NN\\n.,PUNC") and True
        ...     tmpfile.flush()
        ...     _read_csv(tmpfile.name, ',', ['Token']) #doctest: +NORMALIZE_WHITESPACE
        True
              Token
        0      This
        1        is
        2         a
        3       CSV
        4   example
        5         .
    """
    log.debug("Reading columns {} of {} ...".format(columns, filepath))
    if columns is None:
        log.warning("No column names were specified or do not match. The whole file will be processed.")
    return pd.read_csv(filepath, sep=sep, quoting=csv.QUOTE_NONE, usecols=columns)


def _read_txt(filepath):
    """Reads a plain text file based on its path.

    This private function is wrapped in `read_from_pathlist()`.

    Args:
        filepath (str): Path to plain text file.
  
    Returns:
        A ``document`` as str.
    
    Example:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(suffix='.txt') as tmpfile:
        ...     tmpfile.write(b"This is a plain text example.") and True
        ...     tmpfile.flush()
        ...     _read_txt(tmpfile.name)
        True
        'This is a plain text example.'
    """
    log.debug("Reading {} ...".format(filepath))
    with open(filepath, 'r', encoding='utf-8') as document:
        return document.read()


def _read_xml(filepath, xpath_expression):
    """Reads a TEI XML file based on its path.
    
    This private function is wrapped in `read_from_pathlist()`.
    
    Args:
        filepath (str): Path to XML file.
        xpath_expression (str): XPath expressions to match part of the XML file.
    
    Returns:
        Either a ``document`` as str or a list of all parts of the ``document``,
            e. g. chapters of a novel.
    
    Example:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(suffix='.xml') as tmpfile:
        ...     tmpfile.write(b"<text>This is a XML example.</text>") and True
        ...     tmpfile.flush()
        ...     _read_xml(tmpfile.name, '//text')
        True
        'This is a XML example.'
    """
    log.debug("Reading {} matching part or parts of {} ...".format(xpath_expression, filepath))
    ns = dict(tei='http://www.tei-c.org/ns/1.0')
    tree = etree.parse(filepath)
    document = [''.join(element.xpath('.//text()')) for element in tree.xpath(xpath_expression, namespaces=ns)]
    if len(document) == 1:
        return document[0]
    else:
        return document


def _remove_features_from_large_corpus_model(document_term_matrix, type_ids, features):
    """Removes features from large corpus model.

    This private function is wrapped in :func:`remove_features()`.

    Args:
        document_term_matrix (pandas.DataFrame): A document-term matrix.
        type_ids (dict): A dictionary with types as key and identifiers as values.
        features (list): A list of tokens.

    Returns:
        A clean document-term matrix as pandas DataFrame.

    Example:
        >>> document_labels = ['document']
        >>> tokenized_corpus = [['token', 'stopword', 'stopword']]
        >>> document_term_matrix, _, type_ids = create_document_term_matrix(tokenized_corpus, document_labels, large_corpus=True)
        >>> len(_remove_features_from_large_corpus_model(document_term_matrix, type_ids, ['token']))
        1
    """
    features = [token for token in set(type_ids.keys()) if token in features]
    return document_term_matrix.drop([type_ids[token] for token in features], level='type_id')


def _remove_features_from_small_corpus_model(document_term_matrix, features):
    """Removes features from small corpus model.

    This private function is wrapped in :func:`remove_features()`.

    Args:
        document_term_matrix (pandas.DataFrame): A document-term matrix.
        features (list): A list of tokens.

    Returns:
        A clean document-term matrix as pandas DataFrame.

    Example:
        >>> document_labels = ['document']
        >>> tokenized_corpus = [['token', 'stopword', 'stopword']]
        >>> document_term_matrix = create_document_term_matrix(tokenized_corpus, document_labels)
        >>> _remove_features_from_small_corpus_model(document_term_matrix, ['token']) #doctest: +NORMALIZE_WHITESPACE
                  stopword
        document       2.0

    """
    features = [token for token in features if token in document_term_matrix.columns]
    return document_term_matrix.drop(features, axis=1)


def _remove_features_from_tokenized_document(tokenized_document, features):
    """Removes features from a tokenized document.

    This private function is wrapped in :func:`remove_features()`.

    Args:
        tokenized_document (list): The tokenized document to process. This is an iterable of
            tokens.
        features (list): An iterable of tokens.

    Returns:
        A clean tokenized document as list.

    Example:
        >>> tokenized_document = ['token', 'stopword', 'stopword']
        >>> features = ['stopword']
        >>> _remove_features_from_tokenized_document(tokenized_document, features)
        ['token']
    """
    tokenized_document_arr = np.array(tokenized_document)
    features_arr = np.array(features)
    indices = np.where(np.in1d(tokenized_document_arr, features_arr))
    return np.delete(tokenized_document_arr, indices).tolist()


def _stopwords_large_corpus_model(document_term_matrix, type_ids, most_frequent_tokens):
    """Determines stopwords in large corpus model.

    This private function is wrapped in :func:`find_stopwords()`.

    Args:
        document_term_matrix (pandas.DataFrame): A document-term matrix.
        type_ids (dict): A dictionary with types as key and identifiers as values.
        most_frequent_tokens (int, optional): Treshold for most frequent tokens.

    Returns:
        Most frequent tokens in a list.

    Example:
        >>> document_labels = ['document']
        >>> tokenized_corpus = [['hapax', 'stopword', 'stopword']]
        >>> document_term_matrix, _, type_ids = create_document_term_matrix(tokenized_corpus, document_labels, large_corpus=True)
        >>> find_stopwords(document_term_matrix, 1, type_ids)
        ['stopword']
    """
    id2type = {id_: type_ for type_, id_ in type_ids.items()}
    document_term_matrix_collapsed = document_term_matrix.groupby(document_term_matrix.index.get_level_values('type_id')).sum()
    stopwords = document_term_matrix_collapsed[0].nlargest(most_frequent_tokens)
    return [id2type[token] for token in stopwords.index.get_level_values('type_id')]


def _token2id(tokens):
    """Creates a dictionary of tokens as keys and identifier as keys.

    This private function is wrapped in :func:`_create_largecounter()`.

    Args:
        tokens (list): Iterable of tokens.

    Returns:
        A dictionary.

    Example:
        >>> _token2id(['token'])
        {'token': 1}
        >>> _token2id([['token']])
        {'token': 1}
    """
    log.debug("Creating dictionary of tokens as keys and identifier as keys ...")
    if all(isinstance(element, list) for element in tokens):
        tokens = {token for element in tokens for token in element}
    return {token: id_ for id_, token in enumerate(set(tokens), 1)}



"""
Utilizing the Command-Line from Within the Python Environment
*************************************************************

Functions of this module are for **utilizing the command-line**. You can call \
the command-line from within Python using :func:`call_commandline()`, e.g. the \
`DARIAH-DKPro-Wrapper <https://github.com/DARIAH-DE/DARIAH-DKPro-Wrapper>`_ for \
processing and annotating text corpora. Using :class:`Mallet`, you can call the \
NLP-tool MALLET.

Contents
********
    * :func:`call_commandline()` calls based on the elements of a list the command-\
        line.
    * :class:`Mallet` is a class containing methods to call the NLP-tool MALLET.
    * :meth:`call_mallet()` calls MALLET with a specific executable and additional \
        parameteres.
    * :meth:`import_corpus()` imports a text corpus to the specific MALLET corpus \
        format. Uses the executable ``import-dir``.
    * :meth:`train_topics()` creates a topic model with the imported text corpus. \
        Uses the executable ``train-topics``.

"""

import itertools
import logging
import numpy as np
import os
import pandas as pd
import re
import random
from dariah_topics import postprocessing
import shutil
import string
from platform import system
from subprocess import Popen, PIPE
import tempfile

log = logging.getLogger('dariah_topics')


def _decode(std):
    """Decodes the bytes-like output of a subprocess in UTF-8.
    
    This private function is wrapped in :func:`call_commandline()`.
    
    Args:
        std (bytes-like): The ``stdout``  or ``stderr`` (or whatever) of a
            subprocess.
        
    Returns:
        A list of decoded strings.
        
    Example:
        >>> _decode([bytes('This is a test.', encoding='utf-8')])
        ['This is a test.']
    """
    return [line.decode('utf-8').replace('\n', '') for line in std]


def call_commandline(cmd, stdin=None, stdout='pipe', stderr='pipe', communicate=False, logfile=False):
    """Calls the command-line from within Python.
    
    With this function you can call the command-line with a specific command. Each \
    argument has to be an element in a list (``cmd``).
    
    Args:
        cmd (list): A list of command-line arguments.
        stdin (str), optional: Value for stdin. Defaults to None.
        stdout (str), optional: Value for stdout. Defaults to ``pipe``.
        stderr (str), optional: Value for stderr. Defaults to ``pipe``.
        communicate (bool), optioanl: If True, ``stdout`` and ``stderr`` will be
            processed. Defaults to False.
        logfile (bool), optional: If True, a logfile (``commandline.log``) will
            be created. Otherwise ``stdout`` (and ``stderr``, respectively) will
            be printed as logging to the console (level: INFO).
        
    Returns:
        :class:`Popen` object of the subprocess.
        
    Example:
        >>> call_commandline(['python', '-h']) # doctest: +ELLIPSIS
        <subprocess.Popen object at ...>
    """
    if stdin == 'pipe':
        stdin = PIPE
    if stdout == 'pipe':
        stdout = PIPE
    if stderr == 'pipe':
        stderr = PIPE
    
    if not all(isinstance(arg, str) for arg in cmd):
        cmd = [str(arg) for arg in cmd]
    
    log.info("Calling the command-line: {0} ...".format(' '.join(cmd)))

    process = Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr)
    decoded_stderr = _decode(process.stderr)

    if communicate:
        decoded_stderr = _decode(process.stderr)
        decoded_stdout = _decode(process.stdout)
        if logfile:
            log.info("Check commandline.log in '{0}' for logging.".format(os.getcwd()))
            with open('commandline.log', 'w', encoding='utf-8') as file:
                file.write('\n'.join(decoded_stdout))
                file.write('\n'.join(decoded_stderr))
        else:
            for line_stdout in decoded_stdout:
                log.info(line_stdout)
            for line_stdout in decoded_stderr:
                log.info(line_stderr)
    return process


def _check_whitespace(string):
    """Checks if whitespaces are in a string.
    
    This private function is wrapped in :func:`call_mallet()`.
    
    Args:
        string (str): Obviously, a string.
        
    Returns:
        True if whitespaces are **not** in the string, otherwise False.
        
    Example:
        >>> _check_whitespace('nowhitespace')
        True
        >>> _check_whitespace('white space')
        False
    """
    if not re.search(r'\s', str(string)):
        return True
    else:
        return False


def _check_mallet_output(keyword, kwargs=None):
    """Checks if MALLET created output.
    
    This private function is wrapped in :func:`import_tokenized_corpus()` and \
    :func:`train_topics()`.
    
    Args:
        keyword (str): A token, which has to be in ``kwargs.values()``.
        kwargs (dict), optional: Args for the MALLET functions.
    
    Raises:
        OSError, if MALLET did not produce any output files.
    """
    if not 'corpus.mallet' in keyword:
        output_files = [value for arg, value in kwargs.items() if keyword in arg or 'txt' in str(value) or 'xml' in str(value)]
    else:
        output_files = [keyword]

    if not all(os.path.exists(file) for file in output_files):
        raise OSError("MALLET did not produce any output files. Maybe check your args?")


class Mallet:
    """Python wrapper for MALLET.
    
    With this class you can call the command-line tool `MALLET <http://mallet.cs.umass.edu/topics.php>`_ \
    from within Python.
    """
    def __init__(self, executable='mallet', corpus_output=None, logfile=False):
        self.executable = shutil.which(executable)
        if self.executable is None:
            raise FileNotFoundError(("The executable '{0}' could not be found.\n"
                                     "Either place the executable into the $PATH or call "
                                     "{1}(executable='/path/to/mallet')").format(executable, self.__class__.__name__))
        if corpus_output is None:
            self.corpus_output = tempfile.mkdtemp()
        else:
            self.corpus_output = corpus_output
        self.logfile = logfile

    def call_mallet(self, command, **kwargs):
        """Calls the command-line tool MALLET.
        
        With this function you can call `MALLET <http://mallet.cs.umass.edu/topics.php>`_ \
        using a specific ``command`` (e.g. ``train-topics``) and its parameters.
        **Whitespaces (especially for Windows users) are not allowed in paths.**
        
        Args:
            command (str): A MALLET command, this could be ``import-dir`` (load
                the contents of a directory into MALLET instances), ``import-file``
                (load a single file into MALLET instances), ``import-svmlight``
                (load SVMLight format data files into MALLET instances), ``info``
                (get information about MALLET instances), ``train-classifier``
                (train a classifier from MALLET data files), ``classify-dir``
                (classify data from a single file with a saved classifier), ``classify-file``
                (classify the contents of a directory with a saved classifier),
                ``classify-svmlight`` (classify data from a single file in SVMLight
                format), ``train-topics`` (train a topic model from MALLET data
                files), ``infer-topics`` (use a trained topic model to infer topics
                for new documents), ``evaluate-topics`` (estimate the probability
                of new documents under a trained model), ``prune`` (remove features
                based on frequency or information gain), ``split`` (divide data
                into testing, training, and validation portions), ``bulk-load``
                (for big input files, efficiently prune vocabulary and import docs).

        Returns:
            :class:`Popen` object of the MALLET subprocess.
            
        Example:
            >>> import tempfile
            >>> with tempfile.NamedTemporaryFile(suffix='.txt') as tmpfile:
            ...     tmpfile.write(b"This is a plain text example.") and True
            ...     tmpfile.flush()
            ...     Mallet = Mallet(corpus_output='.')
            ...     process = Mallet.call_mallet('import-file', input=tmpfile.name)
            ...     os.path.exists('text.vectors')
            True
            True
            """
        args = [self.executable, command]
        for option, value in kwargs.items():
            args.append('--' + option.replace('_', '-'))
            if value is not None:
                 args.append(value)
                 
        if not all(_check_whitespace(arg) for arg in args):
            raise ValueError("Whitespaces are not allowed in '{0}'".format(args))
            
        if self.logfile:
            communicate = True
        else:
            communicate = False
        
        return call_commandline(args, communicate=communicate, logfile=self.logfile)

    def import_tokenized_corpus(self, tokenized_corpus, document_labels, **kwargs):
        """Creates MALLET corpus model.
        
        With this function you can import a ``tokenized_corpus`` to create the \
        MALLET corpus model. The MALLET command for this step is ``import-dir`` \
        with ``--keep-sequence`` (which is already defined in the function, so \
        you don't have to), but you have the ability to specify all available \
        parameters. The output will be saved in ``output_corpus``.
        
        Args:
            tokenized_corpus (list): Tokenized corpus containing one or more
                iterables containing tokens.
            document_labels (list): Name of each `tokenized_document` in `tokenized_corpus`.
            encoding (str): Character encoding for input file. Defaults to UTF-8.
            token_regex (str): Divides documents into tokens using a regular
                expression (supports Unicode regex). Defaults to \p{L}[\p{L}\p{P}]+\p{L}.
            preserve_case (bool): If False, converts all word features to lowercase.
                Defaults to False.
            remove_stopwords (bool): Ignores a standard list of very common English
                tokens. Defaults to True.
            stoplist (str): Absolute path to plain text stopword list. Defaults to None.
            extra_stopwords (str): Read whitespace-separated words from this file,
                and add them to either the default English stoplist or the list
                specified by ``stoplist``. Defaults to None.
            stop_pattern_file (str): Read regular expressions from a file, one per
                line. Tokens matching these regexps will be removed. Defaults to None.
            skip_header (bool): If True, in each document, remove text occurring
                before a blank line. This is useful for removing email or UseNet
                headers. Defaults to False.
            skip_html (bool): If True, remove text occurring inside <...>, as in
                HTML or SGML. Defaults to False.
            replacement_files (str): Files containing string replacements, one per
                line: 'A B [tab] C' replaces A B with C, 'A B' replaces A B with A_B.
                Defaults to None.
            deletion_files (str): Files containing strings to delete after
                `replacements_files` but before tokenization (i.e. multiword stop
                terms). Defaults to False.
            keep_sequence_bigrams (bool): If True, final data will be a
                FeatureSequenceWithBigrams rather than a FeatureVector. Defaults to False.
            binary_features (bool): If True, features will be binary. Defaults to False.
            save_text_in_source (bool): If True, save original text of document in source.
                Defaults to False.
            print_output (bool): If True, print a representation of the processed data
                to standard output. This option is intended for debugging. Defaults to
                False.

        Returns:
            The absolute path to the created MALLET corpus file.
            
        Example:
            >>> tokenized_corpus = [['this', 'is', 'a', 'tokenized', 'document']]
            >>> document_labels = ['document_label']
            >>> Mallet = Mallet(corpus_output='.')
            >>> mallet_corpus = Mallet.import_tokenized_corpus(tokenized_corpus, document_labels)
            >>> os.path.exists('corpus.mallet')
            True
        """
        corpus_file = os.path.join(self.corpus_output, 'corpus.mallet')
        postprocessing.save_tokenized_corpus(tokenized_corpus, document_labels, self.corpus_output)
        self.call_mallet('import-dir', keep_sequence=None, input=self.corpus_output, output=corpus_file, **kwargs)
        
        _check_mallet_output(os.path.join(self.corpus_output, 'corpus.mallet'))  
        
        return corpus_file

    def train_topics(self, mallet_binary, cleanup=False, **kwargs):
        """Trains LDA model.
        
        With this function you can train a topic model. The MALLET command for \
        this step is ``train-topics`` (which is already defined in the function, \
        so you don't have to), but you have the ability to specify all available \
        parameters.
        
        Args:
            mallet_binary (str): Path to MALLET corpus model.
            cleanup (bool): If True, the directory ``corpus_output`` will be removed
                after modeling.
            input_model (str): The filename from which to read the binary topic
                model.
            input_state (str): The filename from which to read the gzipped Gibbs
                sampling state created by ``output_state``. The original input
                file must be included, using ``input``.
            output_model (str): The filename in which to write the binary topic
                model at the end of the iterations.
            output_state (str): The filename in which to write the Gibbs sampling
                state after at the end of the iterations.
            output_model_interval (int): The number of iterations between writing
                the model (and its Gibbs sampling state) to a binary file. You must
                also set the ``output_model`` to use this option, whose argument
                will be the prefix of the filenames. Default is 0.
            output_state_interval (int): The number of iterations between
                writing the sampling state to a text file. You must also set
                the ``output_state`` to use this option, whose argument will be
                the prefix of the filenames. Default is 0.
            inferencer_filename (str): A topic inferencer applies a previously
                trained topic model to new documents.
            evaluator_filename (str): A held-out likelihood evaluator for new documents.
            output_topic_keys (str): The filename in which to write the top
                words for each topic and any Dirichlet parameters.
            num_top_words (int): The number of most probable words to print for
                each topic after model estimation. Default is 20.
            show_topics_interval (int): The number of iterations between printing
                a brief summary of the topics so far. Default is 50.
            topic_word_weights_file (str): The filename in which to write
                unnormalized weights for every topic and word type.
            word_topic_counts_file (str): The filename in which to write a sparse
                representation of topic-word assignments.
            diagnostics_file (str): The filename in which to write measures of
                topic quality, in XML format.
              Default is null
            xml_topic_report (str): The filename in which to write the top words
                for each topic and any Dirichlet parameters in XML format.
            xml_topic_phrase_report (str): The filename in which to write the top
                words and phrases for each topic and any Dirichlet parameters in
                XML format.
            num_top_docs (int): When writing topic documents with ``output_topic_docs``,
                report this number of top documents. Default is 100
            output_doc_topics (str): The filename in which to write the topic
                proportions per document, at the end of the iterations.
            doc_topics_threshold (float): Do not print topics with proportions less
                than this threshold value within ``output_doc_topics``. Defaults to 0.0.
            num_topics (int): Number of topics. Defaults to 10.
            num_top_words (int): Number of keywords for each topic. Defaults to 10.
            num_interations (int): Number of iterations. Defaults to 1000.
            num_threads (int): Number of threads for parallel training.  Defaults to 1.
            num_icm_iterations (int): Number of iterations of iterated conditional
                modes (topic maximization).  Defaults to 0.
            no_inference (bool): Load a saved model and create a report. Equivalent
                to ``num_iterations = 0``. Defaults to False.
            random_seed (int): Random seed for the Gibbs sampler. Defaults to 0.
            optimize_interval (int): Number of iterations between reestimating
                dirichlet hyperparameters. Defaults to 0.
            optimize_burn_in (int): Number of iterations to run before first
                estimating dirichlet hyperparameters. Defaults to 200.
            use_symmetric_alpha (bool): Only optimize the concentration parameter of
                the prior over document-topic distributions. This may reduce the
                number of very small, poorly estimated topics, but may disperse common
                words over several topics. Defaults to False.
            alpha (float): Sum over topics of smoothing over doc-topic distributions.
                ``alpha_k = [this value] / [num topics]``. Defaults to 5.0.
            beta (float): Smoothing parameter for each topic-word. Defaults to 0.01.
            
        Returns:
            None.
            
        Example:
            >>> tokenized_corpus = [['this', 'is', 'a', 'tokenized', 'document']]
            >>> document_labels = ['document_label']
            >>> Mallet = Mallet(corpus_output='.')
            >>> mallet_corpus = Mallet.import_tokenized_corpus(tokenized_corpus, document_labels)
            >>> mallet_topics = Mallet.train_topics(mallet_corpus,
            ...                                     output_model='model.mallet',
            ...                                     num_iterations=10)
            >>> os.path.exists('model.mallet')
            True
        """
        self.call_mallet('train-topics', input=mallet_binary, **kwargs)
        
        _check_mallet_output('output', kwargs)

        if cleanup:
            shutil.rmtree(self.corpus_output)




"""
Visualizing the Output of LDA Models
************************************

Functions and classes of this module are for visualizing LDA models. This is, \
except one function (:func:`plot_wordcloud`), based on the document-topics \
distribution DataFrame.

Contents
********
    * :func:`plot_wordcloud` plots the top ``n`` words for a specific topic. \
        The higher their weight, the bigger the word.
    * :clas:`PlotDocumentTopics` is basically the core of this module. Construct \
        this class, if you want to plot the content of the document-topics DataFrame.
    * :meth:`static_heatmap` plots a static, :module:`matplotlib`-based heatmap of \
        the document-topics distribution.
    * :meth:`interactive_heatmap` plots an interactive, :module:`bokeh`-based \
        heatmap of the document-topics distribution.
    * :meth:`static_barchart_per_topic` plots a static, :module:`matplotlib`-based  \
        barchart of the document proportions for a specific topic.
    * :meth:`interactive_barchart_per_topic` plots an interactive, :module:`bokeh`-based  \
        barchart of the document proportions for a specific topic.
    * :meth:`static_barchart_per_document` plots a static, :module:`matplotlib`-based  \
        barchart of the topic proportions for a specific document.
    * :meth:`interactive_barchart_per_document` plots an interactive, :module:`bokeh`-based  \
        barchart of the topic proportions for a specific document.
    * :meth:`topic_over_time` plots a static, :module:`matplotlib`-based  \
        line diagram of the development of a topic over time, based on metadata.
    * :meth:`to_file` saves either a :module:`matplotlib` or a :module:`bokeh` figure \
        object to disk.

"""


import logging
from dariah_topics import postprocessing
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np
import os
import pandas as pd
from bokeh.plotting import figure
from bokeh import palettes
from bokeh.models import (
            ColumnDataSource,
            HoverTool,
            LinearColorMapper,
            BasicTicker,
            ColorBar
            )

from collections import Counter


log = logging.getLogger('dariah_topics')


class PlotDocumentTopics:
    """
    Class to visualize document-topic matrix.
    """
    def __init__(self, document_topics):
        self.document_topics = document_topics


    def static_heatmap(self, figsize=(1000 / 96, 600 / 96), dpi=None,
                       labels_fontsize=13, cmap='Blues', ticks_fontsize=12,
                       xlabel='Document', ylabel='Topic', xticks_bottom=0.1,
                       xticks_rotation=50, xticks_ha='right', colorbar=False):
        """Plots a static heatmap.
    
        Args:
            figsize (tuple), optional: Size of the figure in inches. Defaults to
                ``(1000 / 96, 500 / 96)``.
            dpi (int), optional: Dots per inch. Defaults to None.
            labels_fontsize (int), optional: Fontsize of the figure labels. Defaults
                to 13.
            cmap (str), optional: Colormap for the figure. Defaults to ``Blues``.
            ticks_fontsize (int), optional: Fontsize of axis ticks. Defaults to 12.
            xlabel (str), optional: Label of x-axis. Defaults to ``Document``.
            ylabel (str), optional: Label of y-axis. Defaults to ``Topic``.
            xticks_bottom (str), optional: Distance to bottom of x-ticks. Defaults
                to 0.1.
            xticks_rotation (int), optional: Rotation degree of x-ticks. Defaults
                to 50.
            xticks_ha (str), optional: The horizontal alignment of the x-tick labels.
                Defaulst to ``right``.
            colorbar (bool), optional: If True, include colorbar. Defaults to True.

        Returns:
            Figure object.
        """
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        heatmap = ax.pcolor(self.document_topics, cmap=cmap)
        ax.set_xlabel(xlabel, fontsize=labels_fontsize)
        ax.set_ylabel(ylabel, fontsize=labels_fontsize)
        ax.set_xticks(np.arange(self.document_topics.shape[1]) + 0.5)
        ax.set_yticks(np.arange(self.document_topics.shape[0]) + 0.5)
        ax.set_xticklabels(list(self.document_topics.columns), fontsize=ticks_fontsize)
        ax.set_yticklabels(list(self.document_topics.index), fontsize=ticks_fontsize)
        fig.autofmt_xdate(bottom=xticks_bottom, rotation=xticks_rotation, ha=xticks_ha)
        if colorbar:
            cax = ax.imshow(self.document_topics, interpolation='nearest', cmap=cmap)
            cbar = fig.colorbar(cax, ticks=np.arange(0, 1, 0.1))
        return fig
        
    def __static_barchart(self, index, describer, figsize=(11, 7), color='#053967',
                          edgecolor=None, linewidth=None, alpha=None, labels_fontsize=15,
                          ticks_fontsize=14, title=True, title_fontsize=17,
                          dpi=None, transpose_data=False):
        """Plots a static barchart.
    
        Args:
            index Union(int, str): Index of document-topics matrix column or
                name of column.
            describer (str): Describer of what the plot shows, e.g. either document
                or topic.
            title (bool), optional: If True, figure will have a title in the format
                ``describer: index``.
            title_fontsize (int), optional: Fontsize of figure title.
            transpose_data (bool): If True. document-topics matrix will be transposed.
                Defaults to False.
            color (str), optional: Color of the bins. Defaults to ``#053967``.
            edgecolor (str), optional: Color of the bin edges. Defaults to None.
            lindewidth (float), optional: Width of bin lines. Defaults to None.
            alpha (float): Alpha value used for blending. Defaults to None.
            figsize (tuple), optional: Size of the figure in inches. Defaults to
                ``(1000 / 96, 500 / 96)``.
            dpi (int), optional: Dots per inch. Defaults to None.
            labels_fontsize (int), optional: Fontsize of the figure labels. Defaults
                to 15.
            ticks_fontsize (int), optional: Fontsize of axis ticks. Defaults to 14.

        Returns:
            Figure object.
        """
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi) 
        if isinstance(index, int):
            if transpose_data:
                proportions = self.document_topics.T.iloc[index]
            else:
                proportions = self.document_topics.iloc[index]
            if title:
                plot_title = '{0}: {1}'.format(describer, proportions.name)
                ax.set_title(plot_title, fontsize=title_fontsize)
        elif isinstance(index, str):
            if transpose:
                proportions = self.document_topics.T.loc[index]
            else:
                proportions = self.document_topics.loc[index]
            if title:
                plot_title = '{}: {}'.format(describer, index)
                ax.set_title(plot_title, fontsize=title_fontsize)
        else:
            raise ValueError("{} must be int or str.".format(index))
        
        y_axis = np.arange(len(proportions))
        x_axis = proportions
        y_ticks_labels = proportions.index
        ax.barh(y_axis, x_axis, color=color, edgecolor=edgecolor, linewidth=linewidth, alpha=alpha)
        ax.set_xlabel('Proportion', fontsize=labels_fontsize)
        ax.set_ylabel(describer, fontsize=labels_fontsize)
        ax.set_yticks(y_axis)
        ax.set_yticklabels(y_ticks_labels, fontsize=ticks_fontsize)
        ax.tick_params(axis='x', labelsize=ticks_fontsize)
        return fig

    def static_barchart_per_topic(self, **kwargs):
        """Plots a static barchart per topic.
    
        Args:
            index Union(int, str): Index of document-topics matrix column or
                name of column.
            describer (str): Describer of what the plot shows, e.g. either document
                or topic.
            title (bool), optional: If True, figure will have a title in the format
                ``describer: index``.
            title_fontsize (int), optional: Fontsize of figure title.
            transpose_data (bool): If True. document-topics matrix will be transposed.
                Defaults to False.
            color (str), optional: Color of the bins. Defaults to ``#053967``.
            edgecolor (str), optional: Color of the bin edges. Defaults to None.
            lindewidth (float), optional: Width of bin lines. Defaults to None.
            alpha (float): Alpha value used for blending. Defaults to None.
            figsize (tuple), optional: Size of the figure in inches. Defaults to
                ``(1000 / 96, 500 / 96)``.
            dpi (int), optional: Dots per inch. Defaults to None.
            labels_fontsize (int), optional: Fontsize of the figure labels. Defaults
                to 15.
            ticks_fontsize (int), optional: Fontsize of axis ticks. Defaults to 14.

        Returns:
            Figure object.
        """
        return self.__static_barchart(**kwargs)
        
    def static_barchart_per_document(self, **kwargs):
        """Plots a static barchart per document.
    
        Args:
            index Union(int, str): Index of document-topics matrix column or
                name of column.
            describer (str): Describer of what the plot shows, e.g. either document
                or topic.
            title (bool), optional: If True, figure will have a title in the format
                ``describer: index``.
            title_fontsize (int), optional: Fontsize of figure title.
            transpose_data (bool): If True. document-topics matrix will be transposed.
                Defaults to False.
            color (str), optional: Color of the bins. Defaults to ``#053967``.
            edgecolor (str), optional: Color of the bin edges. Defaults to None.
            lindewidth (float), optional: Width of bin lines. Defaults to None.
            alpha (float): Alpha value used for blending. Defaults to None.
            figsize (tuple), optional: Size of the figure in inches. Defaults to
                ``(1000 / 96, 500 / 96)``.
            dpi (int), optional: Dots per inch. Defaults to None.
            labels_fontsize (int), optional: Fontsize of the figure labels. Defaults
                to 15.
            ticks_fontsize (int), optional: Fontsize of axis ticks. Defaults to 14.

        Returns:
            Figure object.
        """
        return self.__static_barchart(transpose_data=True, **kwargs)

    def interactive_heatmap(self, palette=palettes.Blues[9], reverse_palette=True,
                            tools='hover, pan, reset, save, wheel_zoom, zoom_in, zoom_out',
                            width=1000, height=550, x_axis_location='below', toolbar_location='above',
                            sizing_mode='fixed', line_color=None, grid_line_color=None, axis_line_color=None,
                            major_tick_line_color=None, major_label_text_font_size='9pt',
                            major_label_standoff=0, major_label_orientation=3.14/3, colorbar=True):
        """Plots an interactive heatmap.
    
        Args:
            palette (list), optional: A list of color values. Defaults to ``palettes.Blues[9]``.
            reverse_palette (bool), optional: If True, color values of ``palette`` will
                be reversed. Defaults to True.
            tools (str), optional: Tools, which will be includeded. Defaults to ``hover,
                pan, reset, save, wheel_zoom, zoom_in, zoom_out``.
            width (int), optional: Width of the figure. Defaults to 1000.
            height (int), optional: Height of the figure. Defaults to 550.
            x_axis_location (str), optional: Location of the x-axis. Defaults to
                ``below``.
            toolbar_location (str), optional: Location of the toolbar. Defaults to
                ``above``.
            sizing_mode (str), optional: Size fixed or width oriented. Defaults to ``fixed``.
            line_color (str): Color for lines. Defaults to None.
            grid_line_color (str): Color for grid lines. Defaults to None.
            axis_line_color (str): Color for axis lines. Defaults to None.
            major_tick_line_color (str): Color for major tick lines. Defaults to None.
            major_label_text_font_size (str): Font size for major label text. Defaults
                to ``9pt``.
            major_label_standoff (int): Standoff for major labels. Defaults to 0.
            major_label_orientation (float): Orientation for major labels. Defaults
                to ``3.14 / 3``.
            colorbar (bool): If True, colorbar will be included.

        Returns:
            Figure object.
        """        
        if reverse_palette:
            palette = list(reversed(palette))

        x_range = list(self.document_topics.columns)
        y_range = list(self.document_topics.index)
        
        stacked_data = pd.DataFrame(self.document_topics.stack()).reset_index()
        stacked_data.columns = ['Topics', 'Documents', 'Distributions']
        mapper = LinearColorMapper(palette=palette,
                                   low=stacked_data.Distributions.min(),
                                   high=stacked_data.Distributions.max())
        source = ColumnDataSource(stacked_data)
        
        fig = figure(x_range=x_range,
                     y_range=y_range,
                     x_axis_location=x_axis_location,
                     plot_width=width, plot_height=height,
                     tools=tools, toolbar_location=toolbar_location,
                     sizing_mode=sizing_mode,
                     logo=None)
        fig.rect(x='Documents', y='Topics', source=source, width=1, height=1,
                 fill_color={'field': 'Distributions', 'transform': mapper},
                 line_color=line_color)

        fig.grid.grid_line_color = grid_line_color
        fig.axis.axis_line_color = axis_line_color
        fig.axis.major_tick_line_color = major_tick_line_color
        fig.axis.major_label_text_font_size = major_label_text_font_size
        fig.axis.major_label_standoff = major_label_standoff
        fig.xaxis.major_label_orientation = major_label_orientation
        
        if 'hover' in tools:
            fig.select_one(HoverTool).tooltips = [('x-Axis', '@Documents'),
                                                  ('y-Axis', '@Topics'),
                                                  ('Score', '@Distributions')]

        if colorbar:
            feature = ColorBar(color_mapper=mapper, major_label_text_font_size=major_label_text_font_size,
                               ticker=BasicTicker(desired_num_ticks=len(palette)),
                               label_standoff=6, border_line_color=None, location=(0, 0))
            fig.add_layout(feature, 'right')
        return fig

    
    def __interactive_barchart(self, index, describer, tools='hover, pan, reset, save, wheel_zoom, zoom_in, zoom_out',
                                width=1000, height=400, toolbar_location='above',
                                sizing_mode='fixed', line_color=None, grid_line_color=None, axis_line_color=None,
                                major_tick_line_color=None, major_label_text_font_size='9pt',
                                major_label_standoff=0, title=True, bin_height=0.5,
                                transpose_data=False, bar_color='#053967'):
        """Plots an interactive barchart.
    
        Args:
            index Union(int, str): Index of document-topics matrix column or
                name of column.
            describer (str): Describer of what the plot shows, e.g. either document
                or topic.
            bar_color (str), optional: Color of bars. Defaults to ``#053967``.
            transpose_data (bool): If True. document-topics matrix will be transposed.
                Defaults to False.
            title (bool), optional: If True, figure will have a title in the format
                ``describer: index``.
            tools (str), optional: Tools, which will be includeded. Defaults to ``hover,
                pan, reset, save, wheel_zoom, zoom_in, zoom_out``.
            width (int), optional: Width of the figure. Defaults to 1000.
            height (int), optional: Height of the figure. Defaults to 400.
            x_axis_location (str), optional: Location of the x-axis. Defaults to
                ``below``.
            toolbar_location (str), optional: Location of the toolbar. Defaults to
                ``above``.
            sizing_mode (str), optional: Size fixed or width oriented. Defaults to ``fixed``.
            line_color (str): Color for lines. Defaults to None.
            grid_line_color (str): Color for grid lines. Defaults to None.
            axis_line_color (str): Color for axis lines. Defaults to None.
            major_tick_line_color (str): Color for major tick lines. Defaults to None.
            major_label_text_font_size (str): Font size for major label text. Defaults
                to ``9pt``.
            major_label_standoff (int): Standoff for major labels. Defaults to 0.

        Returns:
            Figure object.
        """
        if isinstance(index, int):
            if transpose_data:
                proportions = self.document_topics.T.iloc[index]
            else:
                proportions = self.document_topics.iloc[index]
            if title:
                plot_title = '{}: {}'.format(describer, proportions.name)
        elif isinstance(index, str):
            if transpose_data:
                proportions = self.document_topics.T.loc[index]
            else:
                proportions = self.document_topics.loc[index]
            if title:
                plot_title = '{}: {}'.format(describer, index)
        else:
            raise ValueError("{} must be int or str.".format(index))

        x_axis = proportions
        y_range = list(proportions.index)

        source = ColumnDataSource(dict(Describer=y_range, Proportion=x_axis))

        fig = figure(y_range=y_range, title=plot_title, plot_width=width, plot_height=height,
                   tools=tools, toolbar_location=toolbar_location,
                   sizing_mode=sizing_mode, logo=None)
        fig.hbar(y='Describer', right='Proportion', height=bin_height, source=source,
               line_color=line_color, color=bar_color)

        fig.xgrid.grid_line_color = None
        fig.x_range.start = 0
        fig.grid.grid_line_color = grid_line_color
        fig.axis.axis_line_color = axis_line_color
        fig.axis.major_tick_line_color = major_tick_line_color
        fig.axis.major_label_text_font_size = major_label_text_font_size
        fig.axis.major_label_standoff = major_label_standoff
        
        if 'hover' in tools:
            fig.select_one(HoverTool).tooltips = [('Proportion', '@Proportion')]
        return fig    
    
    def interactive_barchart_per_topic(self, **kwargs):
        """Plots an interactive barchart per topic.
    
        Args:
            index Union(int, str): Index of document-topics matrix column or
                name of column.
            describer (str): Describer of what the plot shows, e.g. either document
                or topic.
            bar_color (str), optional: Color of bars. Defaults to ``#053967``.
            transpose_data (bool): If True. document-topics matrix will be transposed.
                Defaults to False.
            title (bool), optional: If True, figure will have a title in the format
                ``describer: index``.
            tools (str), optional: Tools, which will be includeded. Defaults to ``hover,
                pan, reset, save, wheel_zoom, zoom_in, zoom_out``.
            width (int), optional: Width of the figure. Defaults to 1000.
            height (int), optional: Height of the figure. Defaults to 400.
            x_axis_location (str), optional: Location of the x-axis. Defaults to
                ``below``.
            toolbar_location (str), optional: Location of the toolbar. Defaults to
                ``above``.
            sizing_mode (str), optional: Size fixed or width oriented. Defaults to ``fixed``.
            line_color (str): Color for lines. Defaults to None.
            grid_line_color (str): Color for grid lines. Defaults to None.
            axis_line_color (str): Color for axis lines. Defaults to None.
            major_tick_line_color (str): Color for major tick lines. Defaults to None.
            major_label_text_font_size (str): Font size for major label text. Defaults
                to ``9pt``.
            major_label_standoff (int): Standoff for major labels. Defaults to 0.

        Returns:
            Figure object.
        """
        return self.__interactive_barchart(**kwargs)

    def interactive_barchart_per_document(self, **kwargs):
        """Plots an interactive barchart per document.
    
        Args:
            index Union(int, str): Index of document-topics matrix column or
                name of column.
            describer (str): Describer of what the plot shows, e.g. either document
                or topic.
            bar_color (str), optional: Color of bars. Defaults to ``#053967``.
            transpose_data (bool): If True. document-topics matrix will be transposed.
                Defaults to False.
            title (bool), optional: If True, figure will have a title in the format
                ``describer: index``.
            tools (str), optional: Tools, which will be includeded. Defaults to ``hover,
                pan, reset, save, wheel_zoom, zoom_in, zoom_out``.
            width (int), optional: Width of the figure. Defaults to 1000.
            height (int), optional: Height of the figure. Defaults to 400.
            x_axis_location (str), optional: Location of the x-axis. Defaults to
                ``below``.
            toolbar_location (str), optional: Location of the toolbar. Defaults to
                ``above``.
            sizing_mode (str), optional: Size fixed or width oriented. Defaults to ``fixed``.
            line_color (str): Color for lines. Defaults to None.
            grid_line_color (str): Color for grid lines. Defaults to None.
            axis_line_color (str): Color for axis lines. Defaults to None.
            major_tick_line_color (str): Color for major tick lines. Defaults to None.
            major_label_text_font_size (str): Font size for major label text. Defaults
                to ``9pt``.
            major_label_standoff (int): Standoff for major labels. Defaults to 0.

        Returns:
            Figure object.
        """
        return self.__interactive_barchart(transpose_data=True, **kwargs)

    def topic_over_time(self, metadata_df, threshold=0.1, starttime=1841, endtime=1920):
        """Creates a visualization that shows topics over time.

        Description:
            With this function you can plot topics over time using metadata stored in the documents name.
            Only works with mallet output.

        Args:
            metadata_df(pd.Dataframe()): metadata created by metadata_toolbox
            threshold(float): threshold set to define if a topic in a document is viable
            starttime(int): sets starting point for visualization
            endtime(int): sets ending point for visualization


        Returns: 
            matplotlib plot

        Note: this function is created for a corpus with filenames that looks like:
                1866_ArticleName.txt

        ToDo: make it compatible with gensim output
                Doctest

        """
        years = list(range(starttime, endtime))

        for topiclabel in self.document_topics.index.values:
            topic_over_threshold_per_year = []
            mask = self.document_topics.loc[topiclabel] > threshold
            df = self.document_topics.loc[topiclabel].loc[mask]
            cnt = Counter()
            for filtered_topiclabel in df.index.values:
                year = metadata_df.loc[filtered_topiclabel, 'year']
                print(year)
                cnt[year] += 1
            for year in years:
                topic_over_threshold_per_year.append(cnt[str(year)])
            plt.plot(years, topic_over_threshold_per_year, label=topiclabel)

        plt.xlabel('Year')
        plt.ylabel('count topics over threshold')
        plt.legend()
        # fig.set_size_inches(18.5, 10.5)
        # fig = plt.figure(figsize=(18, 16))
        return plt.gcf().set_size_inches(18.5, 10.5)

    @staticmethod
    def to_file(fig, filename):
        """Saves a figure object to file.
    
        Args:
            fig Union(bokeh.figure, matplotlib.figure): Figure produced by either
                bokeh or matplotlib.
            filename (str): Name of the file with an extension, e.g. ``plot.png``.

        Returns:
            None.
        """
        import matplotlib
        import bokeh
        if isinstance(fig, bokeh.plotting.figure.Figure):
            ext = os.path.splitext(filename)[1]
            if ext == '.png':
                export_png(fig, filename)
            elif ext == '.svg':
                fig.output_backend = 'svg'
                export_svgs(fig, filename)
            elif ext == '.html':
                output_file(filename)
        elif isinstance(fig, matplotlib.figure.Figure):
            fig.savefig(filename)
        return None




#!/usr/bin/env python3

from setuptools import setup, find_packages

PROJECT = 'DARIAH Topic Modeling'
VERSION = '1.0'
REVISION = '1.0.0.dev'
AUTHOR = 'DARIAH-DE Wuerzburg Group'
AUTHOR_EMAIL = 'pielstroem@biozentrum.uni-wuerzburg.de'
URL = 'https://dariah-de.github.io/Topics'

setup(
    name='dariah_topics',
    version=REVISION,
    description=PROJECT,
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    keywords=['topic modeling', 'lda', 'natural language processing', 'digital humanities'],
    packages=find_packages(exclude=['docs', 'test', 'notebooks']),
    install_requires=[
        'pandas>=0.19.2',
        'regex>=2017.01.14',
        'gensim>=0.13.2',
        'lda>=1.0.5',
        'numpy>=1.3',
        'lxml>=3.6.4',
        'matplotlib>=1.5.3',
        'bokeh>=0.12.6',
        'metadata_toolbox'
    ],
    command_options={
        'build_sphinx': {
            'project': ('setup.py', PROJECT),
            'version': ('setup.py', VERSION),
            'release': ('setup.py', REVISION),
        }
    }
)



"""
Measuring and Evaluating Semantic Coherence of Topics
=====================================================

Topic Models generate probability distributions for words out of a collection of
texts, sorting many single word distributions into distinct semantic groups
called _topics_. These topics constitute groups of semantically related words.
This module provides a method to evaluate the topics quantitatively by semantic
coherence.
"""

from itertools import permutations, combinations
import numpy as np
import pandas as pd


class Preparation:
    """
    Preparation for coherence measures.
    """

    def __init__(self, topics, sparse_bow, type_dictionary):
        """
        Creates objects for topics, sparse_bow and type_dictionary.

        Args:
            topics (pd.DataFrame): A DataFrame containing topic keys.
            sparse_bow (pd.DataFrame): A DataFrame containing MultiIndex with
                `doc_id` and `type_id` and word frequencies.
            type_dictionary (dict): A dictionary containing types as key and
                IDs as values.
        """
        self.topics = topics
        self.sparse_bow = sparse_bow
        self.type_dictionary = type_dictionary

    def segment_topics(self, permutation=False):
        """
        Combines or permutes topic keys to bigrams after translating to their IDs.

        Args:
            permutation (bool): True for permutation. If False topic keys will be
                combined. Defaults to False.

        Returns:
            Series containing bigrams.
        """
        bigrams = []
        for topic in self.topics.iterrows():
            topic = [token2bow(token, self.type_dictionary)
                     for token in topic[1]]
            if permutation:
                bigrams.append(list(permutations(topic, 2)))
            else:
                bigrams.append(list(combinations(topic, 2)))
        return pd.Series(bigrams)

    def calculate_occurences(self, bigrams):
        """
        Counts for each token ID all documents containing the ID

        Args:
            bigrams (pd.Series): Series containing bigrams of combined or permuted
                token IDs.

        Returns:
            Series containing document IDs for each token ID.
        """
        bow = self.sparse_bow.reset_index(level=1)['token_id']
        occurences = pd.Series()
        if isinstance(bigrams, set):
            pass
        else:
            keys = set()
            for topic in bigrams:
                for bigram in topic:
                    keys.add(bigram[0])
                    keys.add(bigram[1])
        for key in keys:
            total = set()
            for doc in bow.groupby(level=0):
                if key in doc[1].values:
                    total.add(doc[0])
            occurences[str(key)] = total
        return occurences


class Measures(Preparation):
    """
    Containing PMI measures
    """

    def __init__(self, sparse_bow, type_dictionary):
        """
        Creates objects for sparse_bow and type_dictionary.

        Args:
            topics (pd.DataFrame): A DataFrame containing topic keys.
            sparse_bow (pd.DataFrame): A DataFrame containing MultiIndex with
                `doc_id` and `type_id` and word frequencies.
            type_dictionary (dict): A dictionary containing types as key and
                IDs as values.
        """
        self.type_dictionary = type_dictionary
        self.sparse_bow = sparse_bow

    def pmi_uci(self, pair, occurences, e=0.1, normalize=False):
        """
        Calculates PMI (UCI) for token pair. This variant of PMI is based on
        Newman et al. 2010 Automatic Evaluation of Topic Coherence.

        Args:
            pair (tuple): Tuple containing two tokens, e.g. ('token1', 'token2')
            occurences (pd.Series): Series containing document occurences.
            e (float): Integer to avoid zero division.
            normalize (bool): If True, PMI (UCI) will be normalized. Defaults to
                False.

        Returns:
            Integer.
        """
        n = len(self.sparse_bow.index.levels[0])
        try:
            k1 = occurences[str(pair[0])]
        except KeyError:
            pass
        try:
            k2 = occurences[str(pair[1])]
        except KeyError:
            pass
        try:
            k1k2 = k1.intersection(k2)
            numerator = (len(k1k2) + e) / n
            denominator = ((len(k1) + e) / n) * ((len(k2) + e) / n)
            if normalize:
                return np.log(numerator / denominator) / -np.log(numerator)
            else:
                return np.log(numerator / denominator)
        except UnboundLocalError:
            pass

    def pmi_umass(self, pair, occurences, e=0.1):
        """
        Calculates PMI (UMass) for token pair. This variant of PMI is based on
        Mimno et al. 2011 Optimizing Semantic Coherence in Topic Models.

        Args:
            pair (tuple): Tuple containing two tokens, e.g. ('token1', 'token2')
            occurences (pd.Series): Series containing document occurences.
            e (float): Integer to avoid zero division.

        Returns:
            Integer.
        """
        n = len(self.sparse_bow.count(level=0))
        try:
            k1 = occurences[str(pair[0])]
        except KeyError:
            pass
        try:
            k2 = occurences[str(pair[1])]
        except KeyError:
            pass
        try:
            k1k2 = k1.intersection(k2)
            numerator = (len(k1k2) + e) / n
            denominator = (len(k2) + e) / n
            return np.log(numerator / denominator)
        except UnboundLocalError:
            pass


class Evaluation(Measures):
    def __init__(self, topics, sparse_bow, type_dictionary):
        """
        Creates objects for topics, sparse_bow and type_dictionary.

        Args:
            topics (pd.DataFrame): A DataFrame containing topic keys.
            sparse_bow (pd.DataFrame): A DataFrame containing MultiIndex with
                `doc_id` and `type_id` and word frequencies.
            type_dictionary (dict): A dictionary containing types as key and
                IDs as values.
        """
        self.topics = topics
        self.sparse_bow = sparse_bow
        self.type_dictionary = type_dictionary

    def calculate_umass(self, mean=True, e=0.1):
        """
        Calculates PMI (UMass) for all topic keys in a DataFrame. This variant of
        PMI is based on Mimno et al. 2011 Optimizing Semantic Coherence in Topic Models.

        Args:
            mean (bool): If True, mean will be calculated for each topic, if
                False, median. Defaults to True.
            e (float): Integer to avoid zero division.

        Returns:
            Series with score for each topic.
        """
        scores = []
        N = len(self.topics.T)
        segmented_topics = self.segment_topics()
        occurences = self.calculate_occurences(bigrams=segmented_topics)
        for topic in segmented_topics:
            pmi = []
            for pair in topic:
                pmi.append(self.pmi_umass(
                    pair=pair, occurences=occurences, e=e))
            if mean:
                scores.append((2 / (N * (N - 1))) * np.mean(pmi))
            else:
                scores.append((2 / (N * (N - 1))) * np.median(pmi))
        return pd.Series(scores)

    def calculate_uci(self, mean=True, normalize=False, e=0.1):
        """
        Calculates PMI (UCI) for all topic keys in a DataFrame. This variant of
        PMI is based on Newman et al. 2010 Automatic Evaluation of Topic Coherence.

        Args:
            mean (bool): If True, mean will be calculated for each topic, if
                False, median. Defaults to True.
            normalize (bool): If True, PMI (UCI) will be normalized. Defaults to
                False.
            e (float): Integer to avoid zero division.

        Returns:
            Series with score for each topic.
        """
        scores = []
        N = len(self.topics.T)
        segmented_topics = self.segment_topics(permutation=True)
        occurences = self.calculate_occurences(bigrams=segmented_topics)
        for topic in segmented_topics:
            pmi = []
            for pair in topic:
                pmi.append(self.pmi_uci(
                    pair=pair, occurences=occurences, normalize=normalize, e=e))
            if mean:
                scores.append((2 / (N * (N - 1))) * np.mean(pmi))
            else:
                scores.append((2 / (N * (N - 1))) * np.median(pmi))
        return pd.Series(scores)
